@@ -53,6 +53,10 @@ struct _GoaDaemon
 
   guint config_timeout_id;
   guint credentials_timeout_id;
+
+  uint32_t notification_id;
+  unsigned int notification_signal_id;
+  GPtrArray *accounts_needing_attention;
 };
 
 enum
@@ -94,6 +98,9 @@ static void ensure_credentials_queue_check (GoaDaemon *self);
 
 static void goa_daemon_check_credentials (GoaDaemon *self);
 static void goa_daemon_reload_configuration (GoaDaemon *self);
+static void goa_daemon_send_notification (GoaDaemon *self);
+static void goa_daemon_withdraw_notification (GoaDaemon *self);
+static void goa_daemon_update_notification (GoaDaemon *self);
 
 G_DEFINE_TYPE (GoaDaemon, goa_daemon, G_TYPE_OBJECT);
 
@@ -116,6 +123,13 @@ goa_daemon_finalize (GObject *object)
 {
   GoaDaemon *self = GOA_DAEMON (object);
 
+  if (self->notification_signal_id != 0)
+    {
+      g_dbus_connection_signal_unsubscribe (self->connection, self->notification_signal_id);
+      goa_daemon_withdraw_notification (self);
+    }
+  g_clear_pointer (&self->accounts_needing_attention, g_ptr_array_unref);
+
   if (self->config_timeout_id != 0)
     {
       g_source_remove (self->config_timeout_id);
@@ -304,6 +318,9 @@ goa_daemon_init (GoaDaemon *self)
                            self,
                            G_CONNECT_SWAPPED);
 
+  /* Notifications */
+  self->accounts_needing_attention = g_ptr_array_new_with_free_func (g_object_unref);
+
   self->ensure_credentials_queue = g_queue_new ();
   queue_check_credentials (self);
 }
@@ -1375,6 +1392,9 @@ remove_account_cb (GObject *source_object, GAsyncResult *res, gpointer user_data
   invocation = G_DBUS_METHOD_INVOCATION (data->invocations->data);
   goa_account_complete_remove (account, invocation);
 
+  g_ptr_array_remove (self->accounts_needing_attention, account);
+  goa_daemon_update_notification (self);
+
   g_object_unref (task);
 }
 
@@ -1487,6 +1507,186 @@ on_account_handle_remove (GoaAccount            *account,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+#define NOTIFICATION_ACTION_FMT "('launch-panel', [<('online-accounts', @av [%s])>], @a{sv} {})"
+
+static void
+weak_ref_free (GWeakRef *weak_ref)
+{
+  g_assert (weak_ref != NULL);
+
+  g_weak_ref_clear (weak_ref);
+  g_free (weak_ref);
+}
+
+static void
+on_notification_signal (GDBusConnection *connection,
+                        const char      *sender_name,
+                        const char      *object_path,
+                        const char      *interface_name,
+                        const char      *signal_name,
+                        GVariant        *parameters,
+                        gpointer         user_data)
+{
+  g_autoptr(GoaDaemon) self = g_weak_ref_get ((GWeakRef *) user_data);
+
+  if (self == NULL)
+    return;
+
+  if (g_strcmp0 (signal_name, "ActionInvoked") == 0)
+    {
+      uint32_t id;
+
+      g_variant_get (parameters, "(u&s)", &id, NULL);
+      if (self->notification_id == id)
+        {
+          g_autofree char *target = NULL;
+
+          if (self->accounts_needing_attention->len == 1)
+            {
+              GoaAccount *account = g_ptr_array_index (self->accounts_needing_attention, 0);
+              g_autofree char *account_id = NULL;
+
+              account_id = g_strdup_printf ("<'%s'>", goa_account_get_id (account));
+              target = g_strdup_printf (NOTIFICATION_ACTION_FMT, account_id);
+            }
+          else
+            {
+              target = g_strdup_printf (NOTIFICATION_ACTION_FMT, "");
+            }
+
+          g_dbus_connection_call (self->connection,
+                                  "org.gnome.Settings",
+                                  "/org/gnome/Settings",
+                                  "org.freedesktop.Application",
+                                  "ActivateAction",
+                                  g_variant_new_parsed (target),
+                                  NULL,
+                                  G_DBUS_CALL_FLAGS_NONE,
+                                  -1,
+                                  NULL,
+                                  NULL,
+                                  NULL);
+        }
+    }
+  else if (g_strcmp0 (signal_name, "NotificationClosed") == 0)
+    {
+      uint32_t id;
+
+      g_variant_get (parameters, "(uu)", &id, NULL);
+      if (self->notification_id == id)
+        self->notification_id = 0;
+    }
+}
+
+static void
+goa_daemon_send_notification_cb (GDBusConnection *connection,
+                                 GAsyncResult    *result,
+                                 gpointer         user_data)
+{
+  g_autoptr(GoaDaemon) self = GOA_DAEMON (g_steal_pointer (&user_data));
+  g_autoptr(GVariant) reply = NULL;
+  g_autoptr(GError) error = NULL;
+
+  reply = g_dbus_connection_call_finish (connection, result, &error);
+  if (reply == NULL)
+    {
+      g_warning ("Failed to notify of required account action: %s", error->message);
+      return;
+    }
+
+  g_variant_get (reply, "(u)", &self->notification_id);
+}
+
+static void
+goa_daemon_send_notification (GoaDaemon *self)
+{
+  g_autofree char *message = NULL;
+
+  if (self->notification_signal_id == 0)
+    {
+      GWeakRef *weak_ref;
+
+      weak_ref = g_new0 (GWeakRef, 1);
+      g_weak_ref_init (weak_ref, self);
+
+      self->notification_signal_id =
+        g_dbus_connection_signal_subscribe (self->connection,
+                                            "org.freedesktop.Notifications",
+                                            "org.freedesktop.Notifications",
+                                            NULL, /* NotificationClosed/ActionInvoked */
+                                            "/org/freedesktop/Notifications",
+                                            NULL,
+                                            G_DBUS_SIGNAL_FLAGS_NONE,
+                                            on_notification_signal,
+                                            g_steal_pointer (&weak_ref),
+                                            (GDestroyNotify) weak_ref_free);
+    }
+
+  if (self->accounts_needing_attention->len == 1)
+    {
+      GoaAccount *account = g_ptr_array_index (self->accounts_needing_attention, 0);
+
+      message = g_strdup_printf (_("Failed to sign in to “%s”"),
+                                 goa_account_get_presentation_identity (account));
+    }
+  else
+    {
+      message = g_strdup (_("Failed to sign in to multiple accounts"));
+    }
+
+  g_dbus_connection_call (self->connection,
+                          "org.freedesktop.Notifications",
+                          "/org/freedesktop/Notifications",
+                          "org.freedesktop.Notifications",
+                          "Notify",
+                          g_variant_new ("(susss@as@a{sv}i)",
+                                         _("Online Accounts"),
+                                         self->notification_id,
+                                         "dialog-warning",
+                                         _("Account Action Required"),
+                                         message,
+                                         g_variant_new_parsed ("@as ['default', '']"),
+                                         g_variant_new_parsed ("@a{sv} {}"),
+                                         -1),
+                          NULL, /* reply type */
+                          G_DBUS_CALL_FLAGS_NONE,
+                          -1,
+                          NULL, /* cancellable */
+                          (GAsyncReadyCallback) goa_daemon_send_notification_cb,
+                          g_object_ref (self));
+}
+
+static void
+goa_daemon_withdraw_notification (GoaDaemon *self)
+{
+  if (self->notification_id > 0)
+    {
+      g_dbus_connection_call (self->connection,
+                              "org.freedesktop.Notifications",
+                              "/org/freedesktop/Notifications",
+                              "org.freedesktop.Notifications",
+                              "CloseNotification",
+                              g_variant_new ("(u)", self->notification_id),
+                              NULL,
+                              G_DBUS_CALL_FLAGS_NONE,
+                              -1,
+                              NULL,
+                              NULL,
+                              NULL);
+    }
+}
+
+static void
+goa_daemon_update_notification (GoaDaemon *self)
+{
+  if (self->accounts_needing_attention->len > 0)
+    goa_daemon_send_notification (self);
+  else
+    goa_daemon_withdraw_notification (self);
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static gboolean
 is_authorization_error (GError *error)
 {
@@ -1500,10 +1700,9 @@ is_authorization_error (GError *error)
       if (SOUP_STATUS_IS_CLIENT_ERROR (error->code))
         ret = TRUE;
     }
-  else if (error->domain == GOA_ERROR)
+  else if (g_error_matches (error, GOA_ERROR, GOA_ERROR_NOT_AUTHORIZED))
     {
-      if (error->code == GOA_ERROR_NOT_AUTHORIZED)
-        ret = TRUE;
+      ret = TRUE;
     }
   return ret;
 }
@@ -1568,6 +1767,7 @@ ensure_credentials_queue_collector (GObject *source_object, GAsyncResult *res, g
               g_message ("%s: Setting AttentionNeeded to TRUE because EnsureCredentials() failed with: %s (%s, %d)",
                          g_dbus_object_get_object_path (G_DBUS_OBJECT (data->object)),
                          error->message, g_quark_to_string (error->domain), error->code);
+              g_ptr_array_add (self->accounts_needing_attention, g_object_ref (account));
             }
         }
 
@@ -1583,6 +1783,7 @@ ensure_credentials_queue_collector (GObject *source_object, GAsyncResult *res, g
           g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (account));
           g_message ("%s: Setting AttentionNeeded to FALSE because EnsureCredentials() succeded\n",
                      g_dbus_object_get_object_path (G_DBUS_OBJECT (data->object)));
+          g_ptr_array_remove (self->accounts_needing_attention, account);
         }
 
       ensure_credentials_queue_complete (data->invocations, account, expires_in, NULL);
@@ -1622,7 +1823,10 @@ ensure_credentials_queue_check (GoaDaemon *self)
     goto out;
 
   if (self->ensure_credentials_queue->length == 0)
-    goto out;
+    {
+      goa_daemon_update_notification (self);
+      goto out;
+    }
 
   g_queue_sort (self->ensure_credentials_queue, ensure_credentials_queue_sort, NULL);
 
