@@ -81,6 +81,9 @@ static GoaProviderFeatures
 get_provider_features (GoaProvider *provider)
 {
   return GOA_PROVIDER_FEATURE_BRANDED |
+         GOA_PROVIDER_FEATURE_MAIL |
+         GOA_PROVIDER_FEATURE_CALENDAR |
+         GOA_PROVIDER_FEATURE_CONTACTS |
          GOA_PROVIDER_FEATURE_FILES;
 }
 
@@ -108,7 +111,25 @@ get_redirect_uri (GoaOAuth2Provider *oauth2_provider)
 static const gchar *
 get_scope (GoaOAuth2Provider *oauth2_provider)
 {
-  return "offline_access files.readwrite files.readwrite.all sites.read.all sites.readwrite.all user.read mail.readwrite contacts.readwrite";
+  return "offline_access" \
+         " calendars.readwrite" \
+         " calendars.readwrite.shared" \
+         " contacts.readwrite" \
+         " contacts.readwrite.shared" \
+         " files.readwrite" \
+         " files.readwrite.all" \
+         " mail.readwrite" \
+         " mail.readwrite.shared" \
+         " mail.send" \
+         " mail.send.shared" \
+         " mailboxsettings.read" \
+         " people.read" \
+         " sites.read.all" \
+         " sites.readwrite.all" \
+         " tasks.readwrite" \
+         " tasks.readwrite.shared" \
+         " user.read" \
+         " user.readbasic.all";
 }
 
 static const gchar *
@@ -260,9 +281,13 @@ build_object (GoaProvider        *provider,
   GoaMsGraphProvider *self = GOA_MS_GRAPH_PROVIDER (provider);
   GoaAccount *account = NULL;
   GKeyFile *goa_conf;
+  GoaMail *mail = NULL;
   const gchar  *provider_type;
   const gchar *identity = NULL;
   gboolean files_enabled = FALSE;
+  gboolean calendar_enabled;
+  gboolean contacts_enabled;
+  gboolean mail_enabled;
   gchar *uri_onedrive = NULL;
   gboolean ret = FALSE;
 
@@ -286,6 +311,39 @@ build_object (GoaProvider        *provider,
   uri_onedrive = g_strconcat ("onedrive://", identity, "/", NULL);
   goa_object_skeleton_attach_files (object, uri_onedrive, files_enabled, FALSE);
   g_free (uri_onedrive);
+
+  /* Email */
+  mail = goa_object_get_mail (GOA_OBJECT (object));
+  mail_enabled = goa_util_provider_feature_is_enabled (goa_conf, provider_type, GOA_PROVIDER_FEATURE_MAIL) &&
+                 g_key_file_get_boolean (key_file, group, "MailEnabled", NULL);
+  if (mail_enabled)
+    {
+      if (mail == NULL)
+        {
+          const gchar *email_address;
+
+          email_address = goa_account_get_presentation_identity (account);
+          mail = goa_mail_skeleton_new ();
+          g_object_set (G_OBJECT (mail), "email-address", email_address, NULL);
+          goa_object_skeleton_set_mail (object, mail);
+        }
+    }
+  else
+    {
+      if (mail != NULL)
+        goa_object_skeleton_set_mail (object, NULL);
+    }
+
+  /* Calendar */
+  calendar_enabled = goa_util_provider_feature_is_enabled (goa_conf, provider_type, GOA_PROVIDER_FEATURE_CALENDAR) &&
+                     g_key_file_get_boolean (key_file, group, "CalendarEnabled", NULL);
+  goa_object_skeleton_attach_calendar (object, NULL, calendar_enabled, FALSE);
+
+  /* Contacts */
+  contacts_enabled = goa_util_provider_feature_is_enabled (goa_conf, provider_type, GOA_PROVIDER_FEATURE_CONTACTS) &&
+                     g_key_file_get_boolean (key_file, group, "ContactsEnabled", NULL);
+  goa_object_skeleton_attach_contacts (object, NULL, contacts_enabled, FALSE);
+
   g_clear_pointer (&goa_conf, g_key_file_free);
 
   self->client_id = g_key_file_get_string (key_file, group, "client_id", NULL);
@@ -296,16 +354,32 @@ build_object (GoaProvider        *provider,
   if (just_added)
     {
       goa_account_set_files_disabled (account, !files_enabled);
+      goa_account_set_mail_disabled (account, !mail_enabled);
+      goa_account_set_calendar_disabled (account, !calendar_enabled);
+      goa_account_set_contacts_disabled (account, !contacts_enabled);
 
       g_signal_connect (account,
                         "notify::files-disabled",
                         G_CALLBACK (goa_util_account_notify_property_cb),
                         (gpointer) "FilesEnabled");
+      g_signal_connect (account,
+                        "notify::mail-disabled",
+                        G_CALLBACK (goa_util_account_notify_property_cb),
+                        (gpointer) "MailEnabled");
+      g_signal_connect (account,
+                        "notify::calendar-disabled",
+                        G_CALLBACK (goa_util_account_notify_property_cb),
+                        (gpointer) "CalendarEnabled");
+      g_signal_connect (account,
+                        "notify::contacts-disabled",
+                        G_CALLBACK (goa_util_account_notify_property_cb),
+                        (gpointer) "ContactsEnabled");
     }
 
   ret = TRUE;
 
 out:
+  g_clear_object (&mail);
   g_clear_object (&account);
   return ret;
 }
@@ -318,6 +392,9 @@ add_account_key_values (GoaOAuth2Provider *provider,
   g_assert (builder != NULL);
 
   g_variant_builder_add (builder, "{ss}", "FilesEnabled", "true");
+  g_variant_builder_add (builder, "{ss}", "MailEnabled", "true");
+  g_variant_builder_add (builder, "{ss}", "CalendarEnabled", "true");
+  g_variant_builder_add (builder, "{ss}", "ContactsEnabled", "true");
 
   /* The class getter implementations are used, since they handle fallbacks */
   g_variant_builder_add (builder, "{ss}", "OAuth2AuthorizationUri", goa_oauth2_provider_get_authorization_uri (provider));
@@ -335,6 +412,7 @@ typedef struct
   GoaClient *client;
   GoaObject *object;
 
+  GtkWidget *organization_switch;
   GtkWidget *client_id_entry;
   GtkWidget *tenant_id_entry;
 } AccountData;
@@ -350,11 +428,16 @@ account_data_free (gpointer user_data)
 }
 
 static void
-on_client_id_entry_changed (GtkEditable *editable,
-                            AccountData *data)
+on_entry_changed (GtkEditable *editable,
+                  AccountData *data)
 {
   GoaDialogState state = GOA_DIALOG_READY;
-  const char *client_id;
+  const char *client_id = NULL;
+  const char *tenant_id = NULL;
+
+  tenant_id = gtk_editable_get_text (GTK_EDITABLE (data->tenant_id_entry));
+  if (tenant_id == NULL || *tenant_id == '\0')
+    state = GOA_DIALOG_IDLE;
 
   /* Ensure there is client ID, since there is no build-system default */
   client_id = gtk_editable_get_text (GTK_EDITABLE (data->client_id_entry));
@@ -365,6 +448,17 @@ on_client_id_entry_changed (GtkEditable *editable,
   goa_provider_dialog_set_state (data->dialog, state);
 }
 
+static void
+on_organization_changed (AdwSwitchRow *row,
+                         GParamSpec   *pspec,
+                         AccountData  *data)
+{
+  if (adw_switch_row_get_active (row))
+    on_entry_changed (NULL, data);
+  else
+    goa_provider_dialog_set_state (data->dialog, GOA_DIALOG_READY);
+}
+
 static void
 create_account_details_ui (GoaProvider *provider,
                            AccountData *data,
@@ -379,26 +473,44 @@ create_account_details_ui (GoaProvider *provider,
 
       goa_provider_dialog_add_page (dialog,
                                     NULL, // provider name
-                                    _("Connect to a Microsoft 365 provider to access files"));
+                                    _("Connect to Microsoft 365 to access files stored in OneDrive"));
 
-      group = goa_provider_dialog_add_group (dialog, _("Authorization Details"));
+      group = goa_provider_dialog_add_group (dialog, NULL);
+      data->organization_switch = g_object_new (ADW_TYPE_SWITCH_ROW,
+                                                "title", _("Use Organization Account"),
+                                                "subtitle", _("Connect using details from an organization or developer account"),
+                                                "active", FALSE,
+                                                NULL);
+      adw_preferences_group_add (ADW_PREFERENCES_GROUP (group), data->organization_switch);
+
+      group = goa_provider_dialog_add_group (dialog, NULL);
       /* Translators: See https://learn.microsoft.com/globalization/reference/microsoft-terminology */
-      adw_preferences_group_set_description (ADW_PREFERENCES_GROUP (group), _("A custom Client or Tenant ID may need to be provided depending on the settings for your organization"));
+      adw_preferences_group_set_description (ADW_PREFERENCES_GROUP (group), _("Enter the details provided by your organization"));
 
-      data->client_id_entry = goa_provider_dialog_add_entry (dialog, group, _("_Client ID"));
+      data->client_id_entry = goa_provider_dialog_add_entry (dialog, group, _("_Client ID (Optional)"));
       /* Translators: See https://learn.microsoft.com/globalization/reference/microsoft-terminology */
       data->tenant_id_entry = goa_provider_dialog_add_entry (dialog, group, _("_Tenant ID"));
       goa_provider_dialog_add_description (dialog, data->tenant_id_entry, _("Example ID: 00000000-0000-0000-0000-000000000000"));
 
       button = adw_dialog_get_default_widget (ADW_DIALOG (dialog));
-      gtk_button_set_label (GTK_BUTTON (button), _("_Sign in…"));
+      gtk_button_set_label (GTK_BUTTON (button), _("_Sign In…"));
 
+      g_object_bind_property (data->organization_switch, "active",
+                              group, "visible",
+                              G_BINDING_SYNC_CREATE);
+      g_signal_connect (data->organization_switch,
+                        "notify::active",
+                        G_CALLBACK (on_organization_changed),
+                        data);
       g_signal_connect (data->client_id_entry,
                         "changed",
-                        G_CALLBACK (on_client_id_entry_changed),
+                        G_CALLBACK (on_entry_changed),
                         data);
-
-      on_client_id_entry_changed (GTK_EDITABLE (data->client_id_entry), data);
+      g_signal_connect (data->tenant_id_entry,
+                        "changed",
+                        G_CALLBACK (on_entry_changed),
+                        data);
+      on_organization_changed (ADW_SWITCH_ROW (data->organization_switch), NULL, data);
     }
 }
 
@@ -411,11 +523,14 @@ add_account_parent_cb (GoaProvider  *provider,
   GoaObject *object = NULL;
   GError *error = NULL;
 
+  /* The dialog will already be in the %GOA_DIALOG_DONE state, so the task
+   * is completed manually.
+   */
   object = goa_provider_add_account_finish (provider, result, &error);
   if (object != NULL)
-    goa_provider_task_return_account (task, GOA_OBJECT (object));
+    g_task_return_pointer (task, g_steal_pointer (&object), g_object_unref);
   else
-    goa_provider_task_return_error (task, error);
+    g_task_return_error (task, g_steal_pointer (&error));
 }
 
 static void
@@ -480,7 +595,7 @@ add_account (GoaProvider         *provider,
   g_autoptr(GTask) task = NULL;
 
   data = g_new0 (AccountData, 1);
-  data->dialog = goa_provider_dialog_new (provider, client, parent);
+  data->dialog = goa_provider_dialog_new_full (provider, client, parent, 480, 460);
   data->client = g_object_ref (client);
 
   task = g_task_new (provider, cancellable, callback, user_data);
@@ -513,6 +628,12 @@ goa_ms_graph_provider_init (GoaMsGraphProvider *self)
   self->setup_done = FALSE;
 }
 
+static guint
+get_credentials_generation (GoaProvider *provider)
+{
+  return 1;
+}
+
 static void
 goa_ms_graph_provider_class_init (GoaMsGraphProviderClass *klass)
 {
@@ -527,6 +648,7 @@ goa_ms_graph_provider_class_init (GoaMsGraphProviderClass *klass)
   provider_class->get_provider_features = get_provider_features;
   provider_class->build_object = build_object;
   provider_class->add_account = add_account;
+  provider_class->get_credentials_generation = get_credentials_generation;
 
   oauth2_class = GOA_OAUTH2_PROVIDER_CLASS (klass);
   oauth2_class->get_authorization_uri = get_authorization_uri;
